
import warnings

import os
import json
from pyproj import Transformer

if __name__ == "__main__":
    from coords_dict import COORDS
else:
    from earthnet.coords_dict import COORDS

EXTREME_TILES = ["32UMC", "32UNC", "32UPC", "32UQC"]

def get_coords_from_cube(cubename: str, return_meso: bool = False, ignore_warning = False):
    """

    Get the coordinates for a Cube in Lon-Lat-Grid.

    Args:
        cubename (str): cubename (has format tile_startyear_startmonth_startday_endyear_endmonth_endday_hrxmin_hrxmax_hrymin_hrymax_mesoxmin_mesoxmax_mesoymin_mesoymax.npz)
        return_meso (bool, optional): If True returns also the coordinates for the Meso-scale variables in the cube. Defaults to False.

    Returns:
        tuple: Min-Lon, Min-Lat, Max-Lon, Max-Lat or Min-Lon-HR, Min-Lat-HR, Max-Lon-HR, Max-Lat-HR, Min-Lon-Meso, Min-Lat-Meso, Max-Lon-Meso, Max-Lat-Meso
    """    
    
    if not ignore_warning:
        warnings.warn('Getting coordinates to a cube is experimental. The resulting coordinates on Lon-Lat-Grid will never be pixel perfect. Under certain circumstances, the whole bounding box might shifted by up to 0.02° in either direction. Use with caution. EarthNet2021 does not provide geo-referenced data.')
    
    cubetile,_, _,hr_x_min, hr_x_max, hr_y_min, hr_y_max, meso_x_min, meso_x_max, meso_y_min, meso_y_max = os.path.splitext(cubename)[0].split("_")

    tile = COORDS[cubetile]

    transformer = Transformer.from_crs(tile["EPSG"], 4326, always_xy = True)

    tile_x_min, tile_y_max = transformer.transform(tile["MinLon"],tile["MaxLat"], direction = "INVERSE")

    if cubetile in EXTREME_TILES:
        hr_x_min = int(hr_x_min) + 57
        hr_x_max = int(hr_x_max) + 57

    cube_x_min = tile_x_min + 20 * float(hr_y_min)
    cube_x_max = tile_x_min + 20 * float(hr_y_max)
    cube_y_min = tile_y_max - 20 * float(hr_x_min)
    cube_y_max = tile_y_max - 20 * float(hr_x_max)
    
    cube_lon_min, cube_lat_min = transformer.transform(cube_x_min, cube_y_max)
    cube_lon_max, cube_lat_max = transformer.transform(cube_x_max, cube_y_min)
    
    if return_meso:
        meso_x_min = tile_x_min + 20 * float(meso_x_min)
        meso_x_max = tile_x_min + 20 * float(meso_x_max)
        meso_y_min = tile_y_max - 20 * float(meso_y_min)
        meso_y_max = tile_y_max - 20 * float(meso_y_max)

        meso_lon_min, meso_lat_min = transformer.transform(meso_x_min, meso_y_max)
        meso_lon_max, meso_lat_max = transformer.transform(meso_x_max, meso_y_min)

        return cube_lon_min, cube_lat_min, cube_lon_max, cube_lat_max, meso_lon_min, meso_lat_min, meso_lon_max, meso_lat_max

    else:
        return cube_lon_min, cube_lat_min, cube_lon_max, cube_lat_max

def get_coords_from_tile(tilename: str):
    """
    Get the Coordinates for a Tile in Lon-lat-grid.

    Args:
        tilename (str): 5 Letter MGRS tile

    Returns:
        tuple: Min-Lon, Min-Lat, Max-Lon, Max-Lat
    """    
    tile = COORDS[tilename]
    
    return tile["MinLon"], tile["MinLat"], tile["MaxLon"], tile["MaxLat"]

if __name__ == "__main__":
    import fire
    fire.Fire()





import s3fs
import xarray as xr
from pathlib import Path
from tqdm import tqdm

SPLITS = {
    "earthnet2021x": ["train","iid","ood","extreme","seasonal"]
}

def download(dataset = "earthnet2021x", split = "train", save_directory = "data/", proxy = None, limit = None):
    """Download the recent EarthNet datasets
        
        Before downloading, ensure that you have enough free disk space. We recommend 1 TB.

        Specify the directory `save_directory`, where it should be saved. Then choose, which of the splits you want to download.
        All available splits: 
            - For dataset `"earthnet2021x"`: `["train","iid","ood","extreme","seasonal"]`
        
        You can also give `"all"` to splits to download all splits of a particular dataset.

        Args:
            dataset (str): The dataset you wish to download.
            split (str): A split of the given dataset, can also be `"all"` to download all splits of a given dataset
            save_directory (str): The directory where the data shall be saved in, we recommend data/
            proxy (str, optional): If you need to use a http-proxy to access the internet, you may specify it here.
            limit (int, optional): If you only want to download a certain number of samples, you can set a limit here.
    """  
    if split == "all":
        for split in SPLITS[dataset]:
            download(dataset = dataset, split = split, save_directory=save_directory, proxy = proxy, limit = limit)
    else:
        s3 = s3fs.S3FileSystem(anon=True,
                client_kwargs={
                'endpoint_url': 'https://s3.bgc-jena.mpg.de:9000',
                'region_name': 'thuringia',
                },
                config_kwargs = {
                "proxies": {'http': proxy}
                } if proxy else {}
            )

        print(f"Finding files of {dataset}, split {split} to download.")
        files = s3.find(f"earthnet/{dataset}/{split}")
        print(f"Downloading files of {dataset}, split {split}")
        for file in tqdm(files[:limit] if limit else files):
            savepath = Path(save_directory)/file[9:]
            savepath.parent.mkdir(parents = True, exist_ok = True)
            s3.download(file, str(savepath))
        print(f"Downloaded {dataset}, split {split}.")


def load_minicube(dataset = "earthnet2021x", split = "train", id = "29SND_2018-09-03_2019-01-30_441_569_2745_2873_6_86_42_122", region = None, proxy = None):
    """Load a minicube from a recent EarthNet dataset

        Will give you a minicube loaded from the cloud.

        All available splits: 
            - For dataset `"earthnet2021x"`: `["train","iid","ood","extreme","seasonal"]`

        Args:
            dataset(str): The dataset
            split (str): The split
            id (str): The id of the minicube
            region (str, optional): If you specify the region, downloading will be faster
            proxy (str, optional): If you need to use a http-proxy to access the internet, you may specify it here.
    
    """
    s3 = s3fs.S3FileSystem(anon=True,
            client_kwargs={
            'endpoint_url': 'https://s3.bgc-jena.mpg.de:9000',
            'region_name': 'thuringia',
            },
            config_kwargs = {
            "proxies": {'http': proxy}
            } if proxy else {}
        )
    if region:
        file = f"earthnet/{dataset}/{split}/{region}/{id}.nc"
    else:
        print(f"Searching for {id}...")
        file = s3.glob(f"earthnet/{dataset}/{split}/**/{id}.nc")[0]
        print(f"Found {id}.")
    
    mc = xr.open_dataset(s3.open(file))

    return mc




import numpy as np
import xarray as xr
import pandas as pd
from pathlib import Path
from tqdm import tqdm



def normalized_NSE(targ, pred, name_ndvi_pred = "ndvi_pred"):
    """Compute normalized Nash sutcliffe model efficiency of NDVI for one minicube
    
        The normalized Nash sutcliffe model efficiency scores the predictive skill of a model. It is identical to the most general definition of the coefficient of determination $R^2$.

        We score only on non-cloud observations (determined by the variable `s2_mask` in the target minicube).

        $$$
        nnse = 1 / (2 - nse)
        nse = \frac{\sum_{time} (obs - pred)^2}{\sum_{time} (obs - obsmean)^2}
        $$$

        Further reading: 
        - Normalized NSE https://en.wikipedia.org/wiki/Nash%E2%80%93Sutcliffe_model_efficiency_coefficient
        - NDVI https://de.wikipedia.org/wiki/Normalized_Difference_Vegetation_Index

        Args:
            targ (xr.Dataset): target minicube
            pred (xr.Dataset): prediction minicube, contains `name_ndvi_pred` variable with NDVI predictions during the forecasting period.
            name_ndvi_pred (str, optional): Name of the NDVI prediction variable, defaults to `"ndvi_pred"`.
    """

    pred_start_idx = len(targ.time.isel(time = slice(4,None,5))) - len(pred.time)

    nir = targ.s2_B8A.isel(time = slice(4,None,5)).isel(time = slice(pred_start_idx, None))
    red = targ.s2_B04.isel(time = slice(4,None,5)).isel(time = slice(pred_start_idx, None))
    mask = targ.s2_mask.isel(time = slice(4,None,5)).isel(time = slice(pred_start_idx, None))

    targ_ndvi = ((nir - red) / (nir + red + 1e-8)).where(mask == 0, np.NaN)
    pred_ndvi = pred[name_ndvi_pred]

    nnse = 1 / (2 - (1 - (((targ_ndvi - pred_ndvi)**2).sum("time") / ((targ_ndvi - targ_ndvi.mean("time"))**2).sum("time"))))

    df = xr.Dataset({"NNSE": nnse, "landcover": targ.esawc_lc}).to_dataframe()

    return df.drop(columns="sentinel:product_id", errors = "ignore")

def score_over_dataset(testset_dir, pred_dir, name_ndvi_pred = "ndvi_pred", verbose = True):
    """Compute normalized Nash sutcliffe model efficiency of NDVI for a full dataset

    Args:
        testset_dir (str): directory under which the target minicubes are stored. e.g. `"data/earthnet2021x/test/"`.
        pred_dir (str): directory under which the predictions are stored. e.g. `"preds/my_model/earthnet2021x-test/"`.
        name_ndvi_pred (str, optional): Name of the NDVI prediction variable, defaults to `"ndvi_pred"`.
        verbose (boolean, optional): Set to false to silence this function.
    """

    targetfiles = list(Path(testset_dir).glob("**/*.nc"))

    pred_dir = Path(pred_dir)
    
    if verbose:
        print(f"scoring {testset_dir} against {pred_dir}")

    dfs = []
    for targetfile in tqdm(targetfiles) if verbose else targetfiles:

        cubename = targetfile.name
        region = targetfile.parent.stem

        predfile = pred_dir/region/cubename

        targ = xr.open_dataset(targetfile)
        pred = xr.open_dataset(predfile)

        curr_df = normalized_NSE(targ, pred, name_ndvi_pred=name_ndvi_pred)
        curr_df["id"] = targetfile.stem

        dfs.append(curr_df)

    df = pd.concat(dfs).reset_index()

    tree_score = df[df.landcover == 10.].NNSE.mean()
    shrub_score = df[df.landcover == 20.].NNSE.mean()
    grass_score = df[df.landcover == 30.].NNSE.mean()
    crop_score = df[df.landcover == 40.].NNSE.mean()
    swamp_score = df[df.landcover == 90.].NNSE.mean()
    mangroves_score = df[df.landcover == 95.].NNSE.mean()
    moss_score = df[df.landcover == 100.].NNSE.mean()

    veg_micro_score = df[df.landcover <= 30.].NNSE.mean()
    veg_macro_score = np.nanmean([tree_score, shrub_score, grass_score])

    scores = {
        "veg_macro_score": veg_macro_score,
        "veg_micro_score": veg_micro_score,
        "tree_score": tree_score,
        "shrub_score": shrub_score,
        "grass_score": grass_score,
        "crop_score": crop_score,
        "swamp_score": swamp_score,
        "mangroves_score": mangroves_score,
        "moss_score": moss_score,
        "all_scores": df
    }

    if verbose:
        print("Done!")

    return scores